package dev.fyloz.colorrecipesexplorer.config.security

import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.emergencyMode
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.model.account.User
import dev.fyloz.colorrecipesexplorer.service.users.UserDetailsService
import dev.fyloz.colorrecipesexplorer.service.users.UserService
import org.slf4j.Logger
import org.springframework.boot.context.properties.EnableConfigurationProperties
import org.springframework.context.annotation.Bean
import org.springframework.context.annotation.Configuration
import org.springframework.context.annotation.Lazy
import org.springframework.context.annotation.Profile
import org.springframework.core.env.Environment
import org.springframework.http.HttpMethod
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder
import org.springframework.security.config.annotation.method.configuration.EnableGlobalMethodSecurity
import org.springframework.security.config.annotation.web.builders.HttpSecurity
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter
import org.springframework.security.config.http.SessionCreationPolicy
import org.springframework.security.core.AuthenticationException
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.security.web.AuthenticationEntryPoint
import org.springframework.stereotype.Component
import org.springframework.util.Assert
import org.springframework.web.cors.CorsConfiguration
import org.springframework.web.cors.UrlBasedCorsConfigurationSource
import javax.annotation.PostConstruct
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse

@Configuration
@Profile("!emergency")
@EnableWebSecurity
@EnableGlobalMethodSecurity(prePostEnabled = true)
@EnableConfigurationProperties(CreSecurityProperties::class)
class SecurityConfig(
    private val securityProperties: CreSecurityProperties,
    @Lazy private val userDetailsService: UserDetailsService,
    @Lazy private val userService: UserService,
    private val environment: Environment,
    private val logger: Logger
) : WebSecurityConfigurerAdapter() {
    var debugMode = false

    override fun configure(authBuilder: AuthenticationManagerBuilder) {
        authBuilder.userDetailsService(userDetailsService).passwordEncoder(passwordEncoder())
    }

    @Bean
    fun passwordEncoder() =
        getPasswordEncoder()

    @Bean
    fun corsConfigurationSource() =
        getCorsConfigurationSource()

    @PostConstruct
    fun initWebSecurity() {
        if (emergencyMode) {
            logger.error("Emergency mode is enabled, system users will not be created")
            return
        }

        debugMode = "debug" in environment.activeProfiles
        if (debugMode) logger.warn("Debug mode is enabled, security will be decreased!")

        // Create Root user
        assertRootUserNotNull(securityProperties)
        createSystemUser(
            securityProperties.root!!,
            userService,
            passwordEncoder(),
            "Root",
            "User",
            listOf(Permission.ADMIN)
        )
    }

    override fun configure(http: HttpSecurity) {
        http
            .headers().frameOptions().disable()
            .and()
            .csrf().disable()
            .addFilter(
                JwtAuthenticationFilter(authenticationManager(), securityProperties) {
                    userService.updateLastLoginTime(it)
                }
            )
            .addFilter(
                JwtAuthorizationFilter(securityProperties, authenticationManager()) {
                    userDetailsService.loadUserById(it, false)
                }
            )
            .sessionManagement().sessionCreationPolicy(SessionCreationPolicy.STATELESS)

        if (!debugMode) {
            http
                .authorizeRequests()
                .anyRequest().permitAll()
        } else {
            http
                .cors()
                .and()
                .authorizeRequests()
                .antMatchers("**").permitAll()
        }
    }
}

@Configuration
@Profile("emergency")
@EnableConfigurationProperties(CreSecurityProperties::class)
class EmergencySecurityConfig(
    private val securityProperties: CreSecurityProperties,
    private val userDetailsService: UserDetailsService,
    private val environment: Environment
) : WebSecurityConfigurerAdapter() {
    init {
        emergencyMode = true
    }

    @Bean
    fun corsConfigurationSource() =
        getCorsConfigurationSource()

    @Bean
    fun passwordEncoder() =
        getPasswordEncoder()

    override fun configure(authBuilder: AuthenticationManagerBuilder) {
        authBuilder.userDetailsService(userDetailsService).passwordEncoder(passwordEncoder())
    }

    override fun configure(http: HttpSecurity) {
        val debugMode = "debug" in environment.activeProfiles

        http
            .headers().frameOptions().disable()
            .and()
            .csrf().disable()
            .addFilter(
                JwtAuthenticationFilter(authenticationManager(), securityProperties) { }
            )
            .addFilter(
                JwtAuthorizationFilter(securityProperties, authenticationManager()) {
                    userDetailsService.loadUserById(it, false)
                }
            )
            .sessionManagement().sessionCreationPolicy(SessionCreationPolicy.STATELESS)
            .and()
            .authorizeRequests()
            .antMatchers("**").fullyAuthenticated()
            .antMatchers("/api/login").permitAll()

        if (debugMode) {
            http.cors()
        }
    }
}

@Component
class RestAuthenticationEntryPoint : AuthenticationEntryPoint {
    override fun commence(
        request: HttpServletRequest,
        response: HttpServletResponse,
        authException: AuthenticationException
    ) = response.sendError(HttpServletResponse.SC_UNAUTHORIZED, "Unauthorized")
}

fun createSystemUser(
    credentials: CreSecurityProperties.SystemUserCredentials,
    userService: UserService,
    passwordEncoder: PasswordEncoder,
    firstName: String,
    lastName: String,
    permissions: List<Permission>
) {
    Assert.notNull(credentials.id, "A system user has no identifier defined")
    Assert.notNull(credentials.password, "A system user has no password defined")

    if (!userService.existsById(credentials.id)) {
        userService.save(
            User(
                id = credentials.id,
                firstName = firstName,
                lastName = lastName,
                password = passwordEncoder.encode(credentials.password),
                isSystemUser = true,
                permissions = permissions.toMutableSet()
            )
        )
    }
}

fun getPasswordEncoder() =
    BCryptPasswordEncoder()

fun getCorsConfigurationSource() =
    UrlBasedCorsConfigurationSource().apply {
        registerCorsConfiguration("/**", CorsConfiguration().apply {
            allowedOrigins = listOf("http://localhost:4200")    // Angular development server
            allowedMethods = listOf(
                HttpMethod.GET.name,
                HttpMethod.POST.name,
                HttpMethod.PUT.name,
                HttpMethod.DELETE.name,
                HttpMethod.OPTIONS.name,
                HttpMethod.HEAD.name
            )
            allowCredentials = true
        }.applyPermitDefaultValues())
    }

private fun assertRootUserNotNull(securityProperties: CreSecurityProperties) {
    Assert.notNull(securityProperties.root, "cre.security.root should be defined")
}
