package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeEditUsers
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewUsers
import dev.fyloz.colorrecipesexplorer.model.account.*
import dev.fyloz.colorrecipesexplorer.service.users.GroupService
import dev.fyloz.colorrecipesexplorer.service.users.UserService
import org.springframework.context.annotation.Profile
import org.springframework.http.MediaType
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import java.security.Principal
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.validation.Valid

private const val USER_CONTROLLER_PATH = "api/user"
private const val GROUP_CONTROLLER_PATH = "api/user/group"

@RestController
@RequestMapping(USER_CONTROLLER_PATH)
@Profile("!emergency")
class UserController(private val userService: UserService) {
    @GetMapping
    @PreAuthorizeViewUsers
    fun getAll() =
        ok(userService.getAllForOutput())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(userService.getByIdForOutput(id))

    @GetMapping("current")
    @PreAuthorize("isFullyAuthenticated()")
    fun getCurrent(loggedInUser: Principal?) =
        if (loggedInUser != null)
            ok(
                with(userService) {
                    getById(
                        loggedInUser.name.toLong(),
                        ignoreDefaultGroupUsers = false,
                        ignoreSystemUsers = false
                    ).toOutput()
                }
            )
        else
            forbidden()

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody user: UserSaveDto) =
        created<UserOutputDto>(USER_CONTROLLER_PATH) {
            with(userService) {
                save(user).toOutput()
            }
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody user: UserUpdateDto) =
        noContent {
            userService.update(user)
        }

    @PutMapping("{id}/password", consumes = [MediaType.TEXT_PLAIN_VALUE])
    @PreAuthorizeEditUsers
    fun updatePassword(@PathVariable id: Long, @RequestBody password: String) =
        noContent {
            userService.updatePassword(id, password)
        }

    @PutMapping("{userId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun addPermission(
        @PathVariable userId: Long,
        @PathVariable permission: Permission
    ) = noContent {
        userService.addPermission(userId, permission)
    }

    @DeleteMapping("{userId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun removePermission(
        @PathVariable userId: Long,
        @PathVariable permission: Permission
    ) = noContent {
        userService.removePermission(userId, permission)
    }

    @DeleteMapping("{id}")
    @PreAuthorizeEditUsers
    fun deleteById(@PathVariable id: Long) =
        userService.deleteById(id)
}

@RestController
@RequestMapping(GROUP_CONTROLLER_PATH)
@Profile("!emergency")
class GroupsController(
    private val groupService: GroupService,
    private val userService: UserService
) {
    @GetMapping
    @PreAuthorize("hasAnyAuthority('VIEW_RECIPES', 'VIEW_USERS')")
    fun getAll() =
        ok(groupService.getAllForOutput())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(groupService.getByIdForOutput(id))

    @GetMapping("{id}/users")
    @PreAuthorizeViewUsers
    fun getUsersForGroup(@PathVariable id: Long) =
        ok(with(userService) {
            groupService.getUsersForGroup(id)
                .map { it.toOutput() }
        })

    @PostMapping("default/{groupId}")
    @PreAuthorizeViewUsers
    fun setDefaultGroup(@PathVariable groupId: Long, response: HttpServletResponse) =
        noContent {
            groupService.setResponseDefaultGroup(groupId, response)
        }

    @GetMapping("default")
    @PreAuthorizeViewUsers
    fun getRequestDefaultGroup(request: HttpServletRequest) =
        ok(with(groupService) {
            getRequestDefaultGroup(request).toOutput()
        })

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody group: GroupSaveDto) =
        created<GroupOutputDto>(GROUP_CONTROLLER_PATH) {
            with(groupService) {
                save(group).toOutput()
            }
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody group: GroupUpdateDto) =
        noContent {
            groupService.update(group)
        }

    @DeleteMapping("{id}")
    @PreAuthorizeEditUsers
    fun deleteById(@PathVariable id: Long) =
        noContent {
            groupService.deleteById(id)
        }
}

@RestController
@RequestMapping("api")
@Profile("!emergency")
class LogoutController(private val userService: UserService) {
    @GetMapping("logout")
    @PreAuthorize("isFullyAuthenticated()")
    fun logout(request: HttpServletRequest) =
        ok {
            userService.logout(request)
        }
}
